import keyboard
from random import randint, choice, shuffle
from typing import List, Tuple
import sys
import setup
from field_objects import Player, Anthill, Ant


class Cell:
    '''
    Клетка игрового поля

    Attributes
    ----------
        y : int
            координата y - индекс ряда клеток на поле
        x : int
            координата x - индекс клетки в ряду поля
        content : None | Player | Anthill | Ant
            контент клетки
        image: str
            символ, отображающий контент клетки на поле
    '''

    def __init__(self, y, x):
        self.y = y
        self.x = x
        self.content = None
        self.image = setup.EMPTY

    def draw(self):
        if self.content:
            print(self.content.image, end=' ')
        else:
            print(self.image, end=' ')


class Field:
    '''
    Игровое поле

    Attributes
    ----------
        rows : int
            количество рядов
        cols : int
            количество клеток в ряду
        cells : list
            все клетки
        player : Player
            игрок, создается в центре
        anthills : list
            все муравейники
        ants : list
            все муравьи
        ants_escaped : int
            количество сбежавших муравьев
        messages : int
            сообщения игры о всех действиях за ход, только для разработчика
        turns : int
            количество сделанных ходов
    '''

    def __init__(self, rows: int, cols: int):
        self.rows = rows
        self.cols = cols
        self.cells = self.get_cells()
        self.player = Player(self.rows // 2, self.cols // 2)
        self.cells[self.player.y][self.player.x].content = self.player
        self.anthills = []
        self.make_anthills()
        self.ants = []
        self.ants_escaped = 0
        self.messages = []
        self.turns = 0

    def get_cells(self) -> List[List[Cell]]:
        '''Возвращает новое игровое поле'''
        cells = [
            [Cell(y=row, x=col) for col in range(self.cols)]
            for row in range(self.rows)
        ]
        return cells

    def get_empty_cells(self) -> List[Cell]:
        ''' Возвращает список пустых клеток поля '''
        return [
            cell
            for row in self.cells
            for cell in row
            if cell.content is None
        ]

    def move_player(self) -> bool:
        '''
        Двигает игрока по полю:

        Ждет нажатия клавиш управления.
        Добавляет к текущим координатам игрока дельты dy и dx.
        Игрок не может переместиться в клетку с муравейником.
        Игрок не может переместиться за пределы игрового поля.
        После перемещения игрока в новую клетку контент старой клетки
        меняется на None, а новой - на Player.
        Игрок съест муравья, елсли муравей будет контентом новой клетки.
        Возвращает True, если игрок смог двинуться,
        иначе возвращает False.
        '''

        event = keyboard.read_event(suppress=True)
        if not event.event_type == keyboard.KEY_DOWN:
            return False

        key = event.name

        if key not in setup.KEYS.values():
            return False

        dx, dy = 0, 0
        if key == setup.KEYS['right']:
            dx = 1
        elif key == setup.KEYS['left']:
            dx = -1
        elif key == setup.KEYS['up']:
            dy = -1
        elif key == setup.KEYS['down']:
            dy = 1
        elif key == setup.KEYS['exit']:
            sys.exit()

        this_cell = self.cells[self.player.y][self.player.x]
        new_y = self.player.y + dy
        new_x = self.player.x + dx

        # коллизия с границей поля
        if not self.is_on_field(new_y, new_x):
            return False

        new_cell = self.cells[new_y][new_x]

        # коллизия с муравейником
        if isinstance(new_cell.content, Anthill):
            return False

        # съел муравья
        if isinstance(new_cell.content, Ant):
            self.player.ants_eaten += 1
            self.ants.remove(new_cell.content)

        this_cell.content = None
        new_cell.content = self.player
        self.player.x = new_x
        self.player.y = new_y
        self.turns += 1
        return True

    def count_ants(self) -> int:
        '''возвращает сумму муравьев во всех муравейниках'''
        return sum([anthill.ants_counter for anthill in self.anthills])

    def get_neighbours_coords(self, y: int, x: int) -> List[Tuple]:
        '''
        принимает y и x клетки,
        возвращает список кортежей координат 8 соседних клеток вида [(y, x)]
        '''
        neighbours = []
        for dy in [-1, 0, 1]:
            for dx in [-1, 0, 1]:
                if dy == 0 and dx == 0:
                    continue
                neighbours.append((y + dy, x + dx))
        return neighbours

    def is_on_field(self, y: int, x: int) -> bool:
        '''Возвращает True, если клетка[y][x] в пределах поля'''
        return (y > -1 and y < setup.ROWS) and (x > -1 and x < setup.COLS)

    def make_anthills(self) -> List[Anthill]:
        '''
        Создает от setup.ANTHILLS_MIN до setup.ANTHILLS_MAX вкл муравейников
        в случайных пустых клетках поля
        '''
        anthills_ammount = randint(setup.ANTHILLS_MIN, setup.ANTHILLS_MAX)
        empty_cells = self.get_empty_cells()
        for i in range(anthills_ammount):
            if not empty_cells:
                print('Ошибка! Все мравейники не умещаются на поле!')
                sys.exit()
            anthill_cell = choice(empty_cells)
            anthill = Anthill(anthill_cell.y, anthill_cell.x)
            self.anthills.append(anthill)
            anthill_cell.content = anthill
            empty_cells.remove(anthill_cell)

    def spawn_ants(self) -> None:
        '''
        Спаунит муравьев в клетки поля:
        выбирает место спауна муравья в пределах +- 1 клетки от муравейникаl
        не спаунит муравья на занятые клетки;
        пропускает спаун, если вокруг муравейника нет свободных клеток;
        пропускает спаун, если в муравейнике кончились муравьи.
        '''
        for anthill in self.anthills:
            if not anthill.ants_counter:
                continue

            neighbours_coords = self.get_neighbours_coords(
                anthill.y,
                anthill.x,
            )
            if not neighbours_coords:
                continue

            neighbours = []
            for y, x in neighbours_coords:
                if not self.is_on_field(y, x):
                    continue
                if self.cells[y][x].content:
                    continue
                neighbours.append(self.cells[y][x])

            if not neighbours:
                continue

            ant_cell = choice(neighbours)
            ant = Ant(ant_cell.y, ant_cell.x)
            self.ants.append(ant)
            ant_cell.content = ant
            anthill.ants_counter -= 1
            self.messages.append(
                f'Муравей заспаунился в {ant_cell.y} {ant_cell.x}'
            )

    def move_ants(self) -> None:
        '''
        Двигает муравья на новую случайную из доступных соседних клеток
        Если новая клетка за пределами поля, муравей удаляется - сбежал
        Если новая клетка занята, муравей выбирает другую
        '''
        for ant in self.ants:
            this_cell = self.cells[ant.y][ant.x]
            neighbours_coords = self.get_neighbours_coords(
                ant.y,
                ant.x
            )
            shuffle(neighbours_coords)
            for y, x in neighbours_coords:

                # муравей сбежал
                if not self.is_on_field(y, x):
                    this_cell.content = None
                    self.ants.remove(ant)
                    self.ants_escaped += 1
                    self.messages.append(
                        f'Муравей {ant.y} {ant.x} сбежал в {y} {x}'
                    )
                    break

                new_cell = self.cells[y][x]

                # не спауним в занятую клетку
                if new_cell.content:
                    continue

                this_cell.content = None
                ant.y, ant.x = y, x
                new_cell.content = ant
                self.messages.append(
                    f'Муравей {ant.y} {ant.x} перешел в {y} {x}'
                )
                break

    def draw(self) -> None:
        ''' Рисует игровое поле c декоративной рамкой '''
        print('┌' + '─' * (self.cols * 2 - 1) + '─' + '┐')
        for row in self.cells:
            print('│', end='')
            for cell in row:
                cell.draw()
            print('│')
        print('└' + '─' * (self.cols * 2 - 1) + '─' + '┘')
        print('')  # отступ от легенды, подсказок, статистики
