from os import system
import keyboard
import setup
import assets
from field import Field


class Application:
    '''
    Основная программа

    Attributes
    ----------
        is_running : bool
            контролирует главный цикл программы
        game : Game
            экземпляр игровой сессии
    '''
    def __init__(self) -> None:
        self.is_running = True
        self.game = Game()
        self.run()

    def run(self) -> None:
        '''
        Показывает стартовый экран с картинкой и историей,
        запускает новую игру,
        показывает меню в конце игры
        '''
        self.show_start_screen()
        while self.is_running:
            self.game.start_new_game()
            self.show_end_screen()

    def show_start_screen(self) -> None:
        ''' Выводит на экран ASCII картинку и текст истории '''
        system('cls')
        print(assets.logo)
        print(assets.story)
        input('Нажмите ENTER чтобы начать игру')

    def show_end_screen(self) -> None:
        ''' Выводит на экран меню выбора: новая игра и выход '''
        print('Enter - сыграть еще раз')
        print('Esc - выход')

        while True:
            event = keyboard.read_event(suppress=True)
            if event.event_type == keyboard.KEY_DOWN:
                if event.name == 'esc':
                    self.is_running = False
                    return
                elif event.name == 'enter':
                    return


class Game:
    '''
    Игровая сессия

    Attributes
    ----------
        field: Field
            игровое поле
        is_running: bool
            контролирует главный цикл игровой сессии
    '''

    def __init__(self):
        pass

    def start_new_game(self) -> None:
        '''Запускает новую игровую сессию'''
        self.field = Field(rows=setup.ROWS, cols=setup.COLS)
        self.is_running = True
        self.run()

    def draw_legend(self) -> None:
        '''Выводит на экран легенду'''
        print(
            f'{setup.PLAYER} – игрок, {setup.ANT} – муравей,',
            f'{setup.ANTHILL} – муравейник'
        )
        print()

    def draw_keys(self) -> None:
        '''
        Выводит на экран клавиши управления
        '''
        print(
            f'{setup.KEYS["up"]} – вверх, {setup.KEYS["down"]} – вниз',
            f'{setup.KEYS["left"]} – влево, {setup.KEYS["right"]} – вправо',
            f'{setup.KEYS["exit"]} – выход',
            sep='\n'
        )
        print()

    def draw_statistics(self) -> None:
        '''
        Выводит на экран статистику
        '''
        print(
            f'в муравейниках: {self.field.count_ants()}',
            f'на поле: {len(self.field.ants)}',
            f'съедено: {self.field.player.ants_eaten}',
            f'сбежало: {self.field.ants_escaped}',
            f'cделано ходов: {self.field.turns} ',
            sep='\n'
        )
        print()

    def check_game_over(self) -> None:
        '''
        Устанавливает атрибут is_running в False,
        если на поле и в муравейниках не осталось муравьев
        '''
        self.is_running = self.field.count_ants() or self.field.ants

    def run(self) -> None:
        '''
        Главный цикл игры:
            очищает экран
            очищает лог сообщений
            отрисовывает поле
            выводит на экран дегенду
            выводит на экран клавиши управления
            выводит на экран статистику
            ход игрока
            ход муравьев
            спаун муравьев из муравейников
            проверка, есть ли еще муравьи
        '''
        while self.is_running:
            system('cls')
            self.field.messages.clear()
            self.field.draw()
            self.draw_legend()
            self.draw_keys()
            self.draw_statistics()

            if setup.IS_DEBUG:
                print(*self.field.messages, sep='\n')

            move = self.field.move_player()
            if not move:
                continue

            self.field.move_ants()
            self.field.spawn_ants()
            self.check_game_over()  # может закончить игру

        # конец игры
        system('cls')
        self.field.messages.clear()
        self.field.draw()
        self.draw_statistics()
        print(
            setup.LIGHT_PURPLE
            + 'Муравьи закончились. Конец игры!'
            + setup.END_COLOR
        )


if __name__ == '__main__':
    Application()
